@@ -26,6 +26,7 @@
 #include <grub/types.h>
 #include <grub/fshelp.h>
 #include <grub/deflate.h>
+#include <grub/safemath.h>
 #include <minilzo.h>
 
 #include "xz.h"
@@ -209,7 +210,7 @@ struct grub_fshelp_node
   struct grub_squash_data *data;
   struct grub_squash_inode ino;
   grub_size_t stsize;
-  struct 
+  struct
   {
     grub_disk_addr_t ino_chunk;
     grub_uint16_t ino_offset;
@@ -242,7 +243,7 @@ read_chunk (struct grub_squash_data *data, void *buf, grub_size_t len,
       csize = SQUASH_CHUNK_SIZE - offset;
       if (csize > len)
 	csize = len;
-  
+
       if (grub_le_to_cpu16 (d) & SQUASH_CHUNK_UNCOMPRESSED)
 	{
 	  grub_disk_addr_t a = chunk_start + 2 + offset;
@@ -255,7 +256,7 @@ read_chunk (struct grub_squash_data *data, void *buf, grub_size_t len,
       else
 	{
 	  char *tmp;
-	  grub_size_t bsize = grub_le_to_cpu16 (d) & ~SQUASH_CHUNK_FLAGS; 
+	  grub_size_t bsize = grub_le_to_cpu16 (d) & ~SQUASH_CHUNK_FLAGS;
 	  grub_disk_addr_t a = chunk_start + 2;
 	  tmp = grub_malloc (bsize);
 	  if (!tmp)
@@ -338,7 +339,7 @@ xz_decompress (char *inbuf, grub_size_t insize, grub_off_t off,
   while (len)
     {
       enum xz_ret xzret;
-      
+
       buf.out_pos = 0;
 
       xzret = xz_dec_run (data->xzdec, &buf);
@@ -398,9 +399,9 @@ squash_mount (grub_disk_t disk)
       return NULL;
     }
 
-  err = grub_disk_read (disk, 
+  err = grub_disk_read (disk,
 			grub_le_to_cpu64 (sb.unk1offset)
-			>> GRUB_DISK_SECTOR_BITS, 
+			>> GRUB_DISK_SECTOR_BITS,
 			grub_le_to_cpu64 (sb.unk1offset)
 			& (GRUB_DISK_SECTOR_SIZE - 1), sizeof (frag), &frag);
   if (grub_errno == GRUB_ERR_OUT_OF_RANGE)
@@ -447,7 +448,7 @@ squash_mount (grub_disk_t disk)
     }
 
   data->blksz = grub_le_to_cpu32 (data->sb.block_size);
-  for (data->log2_blksz = 0; 
+  for (data->log2_blksz = 0;
        (1U << data->log2_blksz) < data->blksz;
        data->log2_blksz++);
 
@@ -459,7 +460,17 @@ grub_squash_read_symlink (grub_fshelp_node_t node)
 {
   char *ret;
   grub_err_t err;
-  ret = grub_malloc (grub_le_to_cpu32 (node->ino.symlink.namelen) + 1);
+  grub_size_t sz;
+
+  if (grub_add (grub_le_to_cpu32 (node->ino.symlink.namelen), 1, &sz))
+    {
+      grub_error (GRUB_ERR_OUT_OF_RANGE, N_("overflow is detected"));
+      return NULL;
+    }
+
+  ret = grub_malloc (sz);
+  if (!ret)
+    return NULL;
 
   err = read_chunk (node->data, ret,
 		    grub_le_to_cpu32 (node->ino.symlink.namelen),
@@ -506,11 +517,16 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
 
   {
     grub_fshelp_node_t node;
-    node = grub_malloc (sizeof (*node) + dir->stsize * sizeof (dir->stack[0]));
+    grub_size_t sz;
+
+    if (grub_mul (dir->stsize, sizeof (dir->stack[0]), &sz) ||
+	grub_add (sz, sizeof (*node), &sz))
+      return 0;
+
+    node = grub_malloc (sz);
     if (!node)
       return 0;
-    grub_memcpy (node, dir,
-		 sizeof (*node) + dir->stsize * sizeof (dir->stack[0]));
+    grub_memcpy (node, dir, sz);
     if (hook (".", GRUB_FSHELP_DIR, node, hook_data))
       return 1;
 
@@ -518,12 +534,15 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
       {
 	grub_err_t err;
 
-	node = grub_malloc (sizeof (*node) + dir->stsize * sizeof (dir->stack[0]));
+	if (grub_mul (dir->stsize, sizeof (dir->stack[0]), &sz) ||
+	    grub_add (sz, sizeof (*node), &sz))
+	  return 0;
+
+	node = grub_malloc (sz);
 	if (!node)
 	  return 0;
 
-	grub_memcpy (node, dir,
-		     sizeof (*node) + dir->stsize * sizeof (dir->stack[0]));
+	grub_memcpy (node, dir, sz);
 
 	node->stsize--;
 	err = read_chunk (dir->data, &node->ino, sizeof (node->ino),
@@ -531,7 +550,10 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
 			  + node->stack[node->stsize - 1].ino_chunk,
 			  node->stack[node->stsize - 1].ino_offset);
 	if (err)
-	  return 0;
+	  {
+	    grub_free (node);
+	    return 0;
+	  }
 
 	if (hook ("..", GRUB_FSHELP_DIR, node, hook_data))
 	  return 1;
@@ -557,6 +579,7 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
 	  enum grub_fshelp_filetype filetype = GRUB_FSHELP_REG;
 	  struct grub_squash_dirent di;
 	  struct grub_squash_inode ino;
+	  grub_size_t sz;
 
 	  err = read_chunk (dir->data, &di, sizeof (di),
 			    grub_le_to_cpu64 (dir->data->sb.diroffset)
@@ -580,7 +603,10 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
 			    grub_le_to_cpu64 (dir->data->sb.diroffset)
 			    + chunk, off);
 	  if (err)
-	    return 0;
+	    {
+	      grub_free (buf);
+	      return 0;
+	    }
 
 	  off += grub_le_to_cpu16 (di.namelen) + 1;
 	  buf[grub_le_to_cpu16 (di.namelen) + 1] = 0;
@@ -589,13 +615,22 @@ grub_squash_iterate_dir (grub_fshelp_node_t dir,
 	  if (grub_le_to_cpu16 (di.type) == SQUASH_TYPE_SYMLINK)
 	    filetype = GRUB_FSHELP_SYMLINK;
 
-	  node = grub_malloc (sizeof (*node)
-			      + (dir->stsize + 1) * sizeof (dir->stack[0]));
-	  if (! node)
-	    return 0;
+	  if (grub_add (dir->stsize, 1, &sz) ||
+	      grub_mul (sz, sizeof (dir->stack[0]), &sz) ||
+	      grub_add (sz, sizeof (*node), &sz))
+	    {
+	      grub_free (buf);
+	      return 0;
+	    }
 
-	  grub_memcpy (node, dir,
-		       sizeof (*node) + dir->stsize * sizeof (dir->stack[0]));
+	  node = grub_malloc (sz);
+	  if (! node)
+	    {
+	      grub_free (buf);
+	      return 0;
+	    }
+
+	  grub_memcpy (node, dir, sz - sizeof(dir->stack[0]));
 
 	  node->ino = ino;
 	  node->stack[node->stsize].ino_chunk = grub_le_to_cpu32 (dh.ino_chunk);
@@ -620,7 +655,7 @@ make_root_node (struct grub_squash_data *data, struct grub_fshelp_node *root)
   root->stack[0].ino_chunk = grub_le_to_cpu32 (data->sb.root_ino_chunk);
   root->stack[0].ino_offset = grub_cpu_to_le16 (data->sb.root_ino_offset);
  return read_chunk (data, &root->ino, sizeof (root->ino),
-		    grub_le_to_cpu64 (data->sb.inodeoffset) 
+		    grub_le_to_cpu64 (data->sb.inodeoffset)
 		    + root->stack[0].ino_chunk,
 		    root->stack[0].ino_offset);
 }
@@ -742,11 +777,11 @@ grub_squash_open (struct grub_file *file, const char *name)
 }
 
 static grub_ssize_t
-direct_read (struct grub_squash_data *data, 
+direct_read (struct grub_squash_data *data,
 	     struct grub_squash_cache_inode *ino,
 	     grub_off_t off, char *buf, grub_size_t len)
 {
-  grub_err_t err;
+  grub_err_t err = GRUB_ERR_NONE;
   grub_off_t cumulated_uncompressed_size = 0;
   grub_uint64_t a = 0;
   grub_size_t i;
@@ -859,7 +894,7 @@ direct_read (struct grub_squash_data *data,
 	  grub_free (block);
 	}
       else
-	err = grub_disk_read (data->disk, 
+	err = grub_disk_read (data->disk,
 			      (ino->cumulated_block_sizes[i] + a + boff)
 			      >> GRUB_DISK_SECTOR_BITS,
 			      (ino->cumulated_block_sizes[i] + a + boff)
@@ -923,7 +958,7 @@ grub_squash_read (grub_file_t file, char *buf, grub_size_t len)
     }
   else
     off -= direct_len;
- 
+
   err = read_chunk (data, &frag, sizeof (frag),
 		    data->fragments, sizeof (frag) * fragment);
   if (err)
@@ -934,7 +969,7 @@ grub_squash_read (grub_file_t file, char *buf, grub_size_t len)
     b = grub_le_to_cpu32 (ino->ino.long_file.offset) + off;
   else
     b = grub_le_to_cpu32 (ino->ino.file.offset) + off;
-  
+
   /* FIXME: cache uncompressed chunks.  */
   if (compressed)
     {
@@ -980,7 +1015,7 @@ grub_squash_close (grub_file_t file)
 }
 
 static grub_err_t
-grub_squash_mtime (grub_device_t dev, grub_int32_t *tm)
+grub_squash_mtime (grub_device_t dev, grub_int64_t *tm)
 {
   struct grub_squash_data *data = 0;
 
@@ -990,16 +1025,16 @@ grub_squash_mtime (grub_device_t dev, grub_int32_t *tm)
   *tm = grub_le_to_cpu32 (data->sb.creation_time);
   squash_unmount (data);
   return GRUB_ERR_NONE;
-} 
+}
 
 static struct grub_fs grub_squash_fs =
   {
     .name = "squash4",
-    .dir = grub_squash_dir,
-    .open = grub_squash_open,
-    .read = grub_squash_read,
-    .close = grub_squash_close,
-    .mtime = grub_squash_mtime,
+    .fs_dir = grub_squash_dir,
+    .fs_open = grub_squash_open,
+    .fs_read = grub_squash_read,
+    .fs_close = grub_squash_close,
+    .fs_mtime = grub_squash_mtime,
 #ifdef GRUB_UTIL
     .reserved_first_sector = 0,
     .blocklist_install = 0,
