@@ -27,6 +27,7 @@
 #include <grub/auth.h>
 #include <grub/i18n.h>
 #include <grub/charset.h>
+#include <grub/safemath.h>
 
 enum update_mode
   {
@@ -95,8 +96,8 @@ init_line (struct screen *screen, struct line *linep)
 {
   linep->len = 0;
   linep->max_len = 80;
-  linep->buf = grub_malloc ((linep->max_len + 1) * sizeof (linep->buf[0]));
-  linep->pos = grub_zalloc (screen->nterms * sizeof (linep->pos[0]));
+  linep->buf = grub_calloc (linep->max_len + 1, sizeof (linep->buf[0]));
+  linep->pos = grub_calloc (screen->nterms, sizeof (linep->pos[0]));
   if (! linep->buf || !linep->pos)
     {
       grub_free (linep->buf);
@@ -113,10 +114,18 @@ ensure_space (struct line *linep, int extra)
 {
   if (linep->max_len < linep->len + extra)
     {
-      linep->max_len = 2 * (linep->len + extra);
-      linep->buf = grub_realloc (linep->buf, (linep->max_len + 1) * sizeof (linep->buf[0]));
+      grub_size_t sz0, sz1;
+
+      if (grub_add (linep->len, extra, &sz0) ||
+	  grub_mul (sz0, 2, &sz0) ||
+	  grub_add (sz0, 1, &sz1) ||
+	  grub_mul (sz1, sizeof (linep->buf[0]), &sz1))
+	return 0;
+
+      linep->buf = grub_realloc (linep->buf, sz1);
       if (! linep->buf)
 	return 0;
+      linep->max_len = sz0;
     }
 
   return 1;
@@ -287,7 +296,7 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
 	  pos = linep->pos + (term_screen - screen->terms);
 
 	  if (!*pos)
-	    *pos = grub_zalloc ((linep->len + 1) * sizeof (**pos));
+	    *pos = grub_calloc (linep->len + 1, sizeof (**pos));
 
 	  if (i == region_start || linep == screen->lines + screen->line
 	      || (i > region_start && mode == ALL_LINES))
@@ -324,7 +333,7 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
       while (y < term_screen->geo.num_entries);
 
       /* Draw up and down arrows.  */
-      
+
       if (term_screen->geo.num_entries == 1)
 	{
 	  if (up || down)
@@ -352,12 +361,12 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
       for (i = 0; i < screen->line; i++)
 	y += get_logical_num_lines (screen->lines + i, term_screen);
       if (cpos >= &(screen->lines[screen->line].pos[term_screen - screen->terms])[0])
-	grub_term_gotoxy (term_screen->term, 
+	grub_term_gotoxy (term_screen->term,
 			  (struct grub_term_coordinate) { cpos->x + term_screen->geo.first_entry_x,
 			      cpos->y + y
 			      + term_screen->geo.first_entry_y });
       else
-	grub_term_gotoxy (term_screen->term, 
+	grub_term_gotoxy (term_screen->term,
 			  (struct grub_term_coordinate) { term_screen->geo.first_entry_x,
 			      y + term_screen->geo.first_entry_y });
 
@@ -471,7 +480,7 @@ insert_string (struct screen *screen, const char *s, int update)
 
 	  /* Insert the string.  */
 	  current_linep = screen->lines + screen->line;
-	  unicode_msg = grub_malloc ((p - s) * sizeof (grub_uint32_t));
+	  unicode_msg = grub_calloc (p - s, sizeof (grub_uint32_t));
 
 	  if (!unicode_msg)
 	    return 0;
@@ -650,14 +659,14 @@ backward_char (struct screen *screen, int update)
       linep = screen->lines + screen->line;
 
       screen->column--;
-      screen->column = grub_unicode_get_comb_start (linep->buf, 
+      screen->column = grub_unicode_get_comb_start (linep->buf,
 						    linep->buf + screen->column)
 	- linep->buf;
 
       grub_unicode_aglomerate_comb (screen->lines[screen->line].buf + screen->column,
 				    screen->lines[screen->line].len - screen->column,
 				    &glyph);
-      screen->column = grub_unicode_get_comb_start (linep->buf, 
+      screen->column = grub_unicode_get_comb_start (linep->buf,
 						    linep->buf + screen->column)
 	- linep->buf;
 
@@ -1019,12 +1028,12 @@ complete (struct screen *screen, int continuous, int update)
     return 1;
 
   insert = grub_normal_do_completion (u8, &restore, store_completion);
-  
+
   if (completion_buffer.buf)
     {
       buflen = grub_strlen (completion_buffer.buf);
-      ucs4 = grub_malloc (sizeof (grub_uint32_t) * (buflen + 1));
-      
+      ucs4 = grub_calloc (buflen + 1, sizeof (grub_uint32_t));
+
       if (!ucs4)
 	{
 	  grub_print_error ();
@@ -1268,7 +1277,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
   for (i = 0; i < (unsigned) screen->num_lines; i++)
     {
       grub_free (screen->lines[i].pos);
-      screen->lines[i].pos = grub_zalloc (screen->nterms * sizeof (screen->lines[i].pos[0]));
+      screen->lines[i].pos = grub_calloc (screen->nterms, sizeof (screen->lines[i].pos[0]));
       if (! screen->lines[i].pos)
 	{
 	  grub_print_error ();
@@ -1278,7 +1287,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
 	}
     }
 
-  screen->terms = grub_zalloc (screen->nterms * sizeof (screen->terms[0]));
+  screen->terms = grub_calloc (screen->nterms, sizeof (screen->terms[0]));
   if (!screen->terms)
     {
       grub_print_error ();
@@ -1403,7 +1412,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
 	    goto fail;
 	  break;
 
-	case '\e':
+	case GRUB_TERM_ESC:
 	  destroy_screen (screen);
 	  return;
 
