@@ -28,7 +28,7 @@
 #include <grub/err.h>
 #include <grub/mm.h>
 
-typedef enum 
+typedef enum
   {
     GPG_ERR_NO_ERROR,
     GPG_ERR_BAD_MPI,
@@ -56,7 +56,6 @@ typedef enum
     GPG_ERR_NOT_FOUND,
     GPG_ERR_NOT_IMPLEMENTED,
     GPG_ERR_NOT_SUPPORTED,
-    GPG_ERROR_CFLAGS,
     GPG_ERR_PUBKEY_ALGO,
     GPG_ERR_SELFTEST_FAILED,
     GPG_ERR_TOO_SHORT,
@@ -65,14 +64,15 @@ typedef enum
     GPG_ERR_WRONG_KEY_USAGE,
     GPG_ERR_WRONG_PUBKEY_ALGO,
     GPG_ERR_OUT_OF_MEMORY,
-    GPG_ERR_TOO_LARGE
+    GPG_ERR_TOO_LARGE,
+    GPG_ERR_ENOMEM
   } gpg_err_code_t;
 typedef gpg_err_code_t gpg_error_t;
 typedef gpg_error_t gcry_error_t;
 typedef gpg_err_code_t gcry_err_code_t;
 #define gcry_error_t gcry_err_code_t
 #if 0
-enum gcry_cipher_modes 
+enum gcry_cipher_modes
   {
     GCRY_CIPHER_MODE_NONE   = 0,  /* Not yet specified. */
     GCRY_CIPHER_MODE_ECB    = 1,  /* Electronic codebook. */
@@ -84,6 +84,11 @@ enum gcry_cipher_modes
   };
 #endif
 
+/* Don't rely on this. Check!  */
+#define GRUB_CRYPTO_MAX_MDLEN 64
+#define GRUB_CRYPTO_MAX_CIPHER_BLOCKSIZE 16
+#define GRUB_CRYPTO_MAX_MD_CONTEXT_SIZE 256
+
 /* Type for the cipher_setkey function.  */
 typedef gcry_err_code_t (*gcry_cipher_setkey_t) (void *c,
 						 const unsigned char *key,
@@ -287,9 +292,10 @@ grub_crypto_xor (void *out, const void *in1, const void *in2, grub_size_t size)
     }
   while (size >= sizeof (grub_uint64_t))
     {
+      /* We've already checked that all pointers are aligned.  */
       *(grub_uint64_t *) (void *) outptr
-	= (*(grub_uint64_t *) (void *) in1ptr
-	   ^ *(grub_uint64_t *) (void *) in2ptr);
+	= (*(const grub_uint64_t *) (const void *) in1ptr
+	   ^ *(const grub_uint64_t *) (const void *) in2ptr);
       in1ptr += sizeof (grub_uint64_t);
       in2ptr += sizeof (grub_uint64_t);
       outptr += sizeof (grub_uint64_t);
@@ -314,23 +320,24 @@ grub_crypto_ecb_encrypt (grub_crypto_cipher_handle_t cipher,
 			 void *out, const void *in, grub_size_t size);
 gcry_err_code_t
 grub_crypto_cbc_encrypt (grub_crypto_cipher_handle_t cipher,
-			 void *out, void *in, grub_size_t size,
+			 void *out, const void *in, grub_size_t size,
 			 void *iv_in);
 gcry_err_code_t
 grub_crypto_cbc_decrypt (grub_crypto_cipher_handle_t cipher,
 			 void *out, const void *in, grub_size_t size,
 			 void *iv);
-void 
+void
 grub_cipher_register (gcry_cipher_spec_t *cipher);
 void
 grub_cipher_unregister (gcry_cipher_spec_t *cipher);
-void 
+void
 grub_md_register (gcry_md_spec_t *digest);
-void 
+void
 grub_md_unregister (gcry_md_spec_t *cipher);
 
 extern struct gcry_pk_spec *grub_crypto_pk_dsa;
 extern struct gcry_pk_spec *grub_crypto_pk_ecdsa;
+extern struct gcry_pk_spec *grub_crypto_pk_ecdh;
 extern struct gcry_pk_spec *grub_crypto_pk_rsa;
 
 void
@@ -400,16 +407,16 @@ void _gcry_assert_failed (const char *expr, const char *file, int line,
                           const char *func) __attribute__ ((noreturn));
 
 void _gcry_burn_stack (int bytes);
-void _gcry_log_error( const char *fmt, ... )  __attribute__ ((format (printf, 1, 2)));
-void _gcry_log_bug( const char *fmt, ... )  __attribute__ ((format (printf, 1, 2)));
-void _gcry_log_printf( const char *fmt, ... )  __attribute__ ((format (printf, 1, 2)));
-void
-_gcry_check_heap (const void *a __attribute__ ((unused)));
+void _gcry_log_error( const char *fmt, ... )  __attribute__ ((format (__printf__, 1, 2)));
 
 
 #ifdef GRUB_UTIL
 void grub_gcry_init_all (void);
 void grub_gcry_fini_all (void);
+
+int
+grub_get_random (void *out, grub_size_t len);
+
 #endif
 
 #endif
