@@ -20,6 +20,7 @@
 #define GRUB_PROCFS_HEADER	1
 
 #include <grub/list.h>
+#include <grub/types.h>
 
 struct grub_procfs_entry
 {
@@ -27,7 +28,7 @@ struct grub_procfs_entry
   struct grub_procfs_entry **prev;
 
   const char *name;
-  char * (*get_contents) (void);
+  char * (*get_contents) (grub_size_t *sz);
 };
 
 extern struct grub_procfs_entry *grub_procfs_entries;
