@@ -273,7 +273,7 @@ struct grub_video_edid_info
 
   grub_uint8_t extension_flag;
   grub_uint8_t checksum;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 typedef enum grub_video_driver_id
   {
@@ -290,7 +290,9 @@ typedef enum grub_video_driver_id
     GRUB_VIDEO_DRIVER_RADEON_FULOONG2E,
     GRUB_VIDEO_DRIVER_COREBOOT,
     GRUB_VIDEO_DRIVER_IEEE1275,
-    GRUB_VIDEO_ADAPTER_CAPTURE
+    GRUB_VIDEO_ADAPTER_CAPTURE,
+    GRUB_VIDEO_DRIVER_XEN,
+    GRUB_VIDEO_DRIVER_RADEON_YEELOONG3A
   } grub_video_driver_id_t;
 
 typedef enum grub_video_adapter_prio
@@ -301,6 +303,11 @@ typedef enum grub_video_adapter_prio
     GRUB_VIDEO_ADAPTER_PRIO_NATIVE = 100
   } grub_video_adapter_prio_t;
 
+typedef enum grub_video_area_status
+  {
+    GRUB_VIDEO_AREA_DISABLED,
+    GRUB_VIDEO_AREA_ENABLED
+  } grub_video_area_status_t;
 
 struct grub_video_adapter
 {
@@ -341,6 +348,16 @@ struct grub_video_adapter
   grub_err_t (*get_viewport) (unsigned int *x, unsigned int *y,
                               unsigned int *width, unsigned int *height);
 
+  grub_err_t (*set_region) (unsigned int x, unsigned int y,
+                            unsigned int width, unsigned int height);
+
+  grub_err_t (*get_region) (unsigned int *x, unsigned int *y,
+                            unsigned int *width, unsigned int *height);
+
+  grub_err_t (*set_area_status) (grub_video_area_status_t area_status);
+
+  grub_err_t (*get_area_status) (grub_video_area_status_t *area_status);
+
   grub_video_color_t (*map_color) (grub_uint32_t color_name);
 
   grub_video_color_t (*map_rgb) (grub_uint8_t red, grub_uint8_t green,
@@ -396,10 +413,14 @@ static inline void
 grub_video_register (grub_video_adapter_t adapter)
 {
   grub_video_adapter_t *p;
-  for (p = &grub_video_adapter_list; *p && (*p)->prio > adapter->prio; 
+  for (p = &grub_video_adapter_list; *p && (*p)->prio > adapter->prio;
        p = &((*p)->next));
   adapter->next = *p;
   *p = adapter;
+
+  adapter->prev = p;
+  if (adapter->next)
+    adapter->next->prev = &adapter->next;
 }
 #endif
 
@@ -443,6 +464,22 @@ grub_err_t EXPORT_FUNC (grub_video_get_viewport) (unsigned int *x,
 						  unsigned int *width,
 						  unsigned int *height);
 
+grub_err_t EXPORT_FUNC (grub_video_set_region) (unsigned int x,
+                                                unsigned int y,
+                                                unsigned int width,
+                                                unsigned int height);
+
+grub_err_t EXPORT_FUNC (grub_video_get_region) (unsigned int *x,
+                                                unsigned int *y,
+                                                unsigned int *width,
+                                                unsigned int *height);
+
+grub_err_t EXPORT_FUNC (grub_video_set_area_status)
+    (grub_video_area_status_t area_status);
+
+grub_err_t EXPORT_FUNC (grub_video_get_area_status)
+    (grub_video_area_status_t *area_status);
+
 grub_video_color_t EXPORT_FUNC (grub_video_map_color) (grub_uint32_t color_name);
 
 grub_video_color_t EXPORT_FUNC (grub_video_map_rgb) (grub_uint8_t red,
@@ -460,7 +497,7 @@ grub_err_t EXPORT_FUNC (grub_video_unmap_color) (grub_video_color_t color,
 						 grub_uint8_t *blue,
 						 grub_uint8_t *alpha);
 
-grub_err_t EXPORT_FUNC (grub_video_fill_rect) (grub_video_color_t color, 
+grub_err_t EXPORT_FUNC (grub_video_fill_rect) (grub_video_color_t color,
 					       int x, int y,
 					       unsigned int width,
 					       unsigned int height);
@@ -515,7 +552,7 @@ grub_video_check_mode_flag (grub_video_mode_type_t flags,
 grub_video_driver_id_t EXPORT_FUNC (grub_video_get_driver_id) (void);
 
 static __inline grub_video_rgba_color_t
-grub_video_rgba_color_rgb (int r, int g, int b)
+grub_video_rgba_color_rgb (grub_uint8_t r, grub_uint8_t g, grub_uint8_t b)
 {
   grub_video_rgba_color_t c;
   c.red = r;
@@ -531,12 +568,6 @@ grub_video_map_rgba_color (grub_video_rgba_color_t c)
   return grub_video_map_rgba (c.red, c.green, c.blue, c.alpha);
 }
 
-int EXPORT_FUNC (grub_video_get_named_color) (const char *name,
-					      grub_video_rgba_color_t *color);
-
-grub_err_t EXPORT_FUNC (grub_video_parse_color) (const char *s,
-						 grub_video_rgba_color_t *color);
-
 #ifndef GRUB_MACHINE_EMU
 extern void grub_font_init (void);
 extern void grub_font_fini (void);
@@ -548,6 +579,8 @@ extern void grub_video_sis315pro_init (void);
 extern void grub_video_radeon_fuloong2e_init (void);
 extern void grub_video_sis315pro_fini (void);
 extern void grub_video_radeon_fuloong2e_fini (void);
+extern void grub_video_radeon_yeeloong3a_init (void);
+extern void grub_video_radeon_yeeloong3a_fini (void);
 #endif
 
 void
