@@ -120,9 +120,9 @@ read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len, void
       return;
     }
 
-  if (uncompress == 0)
-    grub_file_filter_disable_compression ();
-  file = grub_file_open (pathname);
+  file = grub_file_open (pathname, ((uncompress == 0)
+				    ? GRUB_FILE_TYPE_NO_DECOMPRESS : GRUB_FILE_TYPE_NONE)
+			 | GRUB_FILE_TYPE_FSTEST);
   if (!file)
     {
       grub_util_error (_("cannot open `%s': %s"), pathname,
@@ -199,7 +199,7 @@ cp_hook (grub_off_t ofs, char *buf, int len, void *_ctx)
 static void
 cmd_cp (char *src, const char *dest)
 {
-  struct cp_hook_ctx ctx = 
+  struct cp_hook_ctx ctx =
     {
       .dest = dest
     };
@@ -300,9 +300,15 @@ cmd_cmp (char *src, char *dest)
 	  strcpy (ptr, entry->d_name);
 
 	  if (grub_util_is_special_file (destnew))
-	    continue;
+	    {
+	      free (srcnew);
+	      free (destnew);
+	      continue;
+	    }
 
 	  cmd_cmp (srcnew, destnew);
+	  free (srcnew);
+	  free (destnew);
 	}
       grub_util_fd_closedir (dir);
       return;
@@ -349,7 +355,7 @@ static int
 crc_hook (grub_off_t ofs, char *buf, int len, void *crc_ctx)
 {
   (void) ofs;
-  
+
   GRUB_MD_CRC32->write(crc_ctx, buf, len);
   return 0;
 }
@@ -470,9 +476,9 @@ fstest (int n)
 	fs = grub_fs_probe (dev);
 	if (!fs)
 	  grub_util_error ("%s", grub_errmsg);
-	if (!fs->uuid)
+	if (!fs->fs_uuid)
 	  grub_util_error ("%s", _("couldn't retrieve UUID"));
-	if (fs->uuid (dev, &uuid))
+	if (fs->fs_uuid (dev, &uuid))
 	  grub_util_error ("%s", grub_errmsg);
 	if (!uuid)
 	  grub_util_error ("%s", _("couldn't retrieve UUID"));
@@ -483,7 +489,7 @@ fstest (int n)
 	grub_device_close (dev);
       }
     }
-    
+
   for (i = 0; i < num_disks; i++)
     {
       char *argv[2];
@@ -512,7 +518,7 @@ static struct argp_option options[] = {
   {N_("crc FILE"), 0, 0     , OPTION_DOC, N_("Get crc32 checksum of FILE."), 1},
   {N_("blocklist FILE"), 0, 0, OPTION_DOC, N_("Display blocklist of FILE."), 1},
   {N_("xnu_uuid DEVICE"), 0, 0, OPTION_DOC, N_("Compute XNU UUID of the device."), 1},
-  
+
   {"root",      'r', N_("DEVICE_NAME"), 0, N_("Set root device."),                 2},
   {"skip",      's', N_("NUM"),           0, N_("Skip N bytes from output file."),   2},
   {"length",    'n', N_("NUM"),           0, N_("Handle N bytes in output file."),   2},
@@ -535,10 +541,10 @@ print_version (FILE *stream, struct argp_state *state)
 }
 void (*argp_program_version_hook) (FILE *, struct argp_state *) = print_version;
 
-static error_t 
+static error_t
 argp_parser (int key, char *arg, struct argp_state *state)
 {
-  char *p;
+  const char *p;
 
   switch (key)
     {
@@ -549,7 +555,7 @@ argp_parser (int key, char *arg, struct argp_state *state)
     case 'K':
       if (strcmp (arg, "prompt") == 0)
 	{
-	  char buf[1024];	  
+	  char buf[1024];
 	  grub_puts_ (N_("Enter ZFS password: "));
 	  if (grub_password_get (buf, 1023))
 	    {
@@ -650,7 +656,7 @@ argp_parser (int key, char *arg, struct argp_state *state)
   if (args_count < num_disks)
     {
       if (args_count == 0)
-	images = xmalloc (num_disks * sizeof (images[0]));
+	images = xcalloc (num_disks, sizeof (images[0]));
       images[args_count] = grub_canonicalize_file_name (arg);
       args_count++;
       return 0;
@@ -722,7 +728,7 @@ argp_parser (int key, char *arg, struct argp_state *state)
 
 struct argp argp = {
   options, argp_parser, N_("IMAGE_PATH COMMANDS"),
-  N_("Debug tool for filesystem driver."), 
+  N_("Debug tool for filesystem driver."),
   NULL, NULL, NULL
 };
 
@@ -734,7 +740,7 @@ main (int argc, char *argv[])
 
   grub_util_host_init (&argc, &argv);
 
-  args = xmalloc (argc * sizeof (args[0]));
+  args = xcalloc (argc, sizeof (args[0]));
 
   argp_parse (&argp, argc, argv, 0, 0, 0);
 
